#
#   Imandra Inc.
#
#   test_formalization.py
#

import unittest


class TestFormalization(unittest.TestCase):
    """
    Test formalization
    """

    def test_single_model(self):
        """
        """

        pass



if __name__ == "__main__":
    unittest.main()