"""Analysis package for advanced mathematical computations."""

from .numerical import derivatives, integration
from .optimization import root_finding

__all__ = ["derivatives", "integration", "root_finding"]